/*
 * match.c - Simple packet matcher using tccext
 *
 * Written 2001 by Werner Almesberger
 * Copyright 2001 Network Robots
 */


#include <inttypes.h>
#include <stdlib.h>
#include <stdio.h>

#include "tccext.h"


#define _S(x) #x
#define S(x) _S(x)


static uint32_t extract_bits(uint8_t *data,int start,int bits)
{
    uint32_t res = 0;

    if ((bits+(start & 7)) > 32) abort();
    data += start >> 3;
    start &= 7;
    res = *data++ & ((1 << (8-start))-1);
    bits -= 8-start;
    while (bits > 0) {
	res = (res << 8) | *data++;
	bits -= 8;
    }
    if (bits) res >>= -bits;
    return res;
}


static int offset(const TCCEXT_OFFSET *group,uint8_t *raw)
{
    if (!group) return 0;
    return offset(group->base,raw)+
      (extract_bits(raw,offset(group->field.offset_group,raw)+
      group->field.offset,group->field.length) << group->shift_left);
}


static int match_1(const TCCEXT_MATCH *m,uint8_t *raw)
{
    return extract_bits(raw,offset(m->field.offset_group,raw)+m->field.offset,
      m->field.length) == extract_bits(m->data,(-m->field.length) & 7,
      m->field.length);
}


int match(uint8_t *raw); /* extern in tcsim */

int match(uint8_t *raw)
{
    static TCCEXT_CONTEXT *context;
    TCCEXT_RULE *r;

    if (!context) {
	FILE *file;

	file = fopen("_" S(NAME) ".out","r");
	if (!file) {
	    perror("_" S(NAME) ".out");
	    exit(1);
	}
	context = tccext_parse(file,NULL);
    }
    for (r = context->rules; r; r = r->next) {
	TCCEXT_MATCH *m;

	for (m = r->matches; m; m = m->next)
	    if (!match_1(m,raw)) break;
	if (!m) return r->action;
    }
    return 0; /* @@@ change once a default value is selected */
}
